// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include <functional>	 // for bind
#include <thread>

#include "Actor.hpp"
#include "BinaryPrivateSemaphore.hpp"
#include "emper-common.h"

class SignallingActor : public Actor<unsigned int> {
 private:
	BinaryPrivateSemaphore& bps;

 protected:
	void receive(UNUSED_ARG unsigned int signalCount) override { bps.signal(); }

 public:
	SignallingActor(BinaryPrivateSemaphore& bps) : bps(bps){};
};

void emperTest() {
	BinaryPrivateSemaphore bps;
	SignallingActor signallingActor(bps);
	signallingActor.start();

	// TODO: Use std::jthread once EMPER uses C++20.
	std::thread signallingThread([&] { signallingActor.tellFromAnywhere(1); });

	bps.wait();

	// TODO: Remove this once we use std::jthread when EMPER uses C++20.
	signallingThread.join();
}
